#include "MeshDrawer.h"
#include "MeshDrawerCoreVisitor.h"

namespace X3DTK {
namespace MESH {

MeshDrawer::MeshDrawer()
: X3DOnePassProcessor(), drawPoints(false), drawEdges(true), drawFaces(true)
{
  globalVariables = GVManager::getInstanceOf<MeshDrawerGlobalVariables>();

  graphTraversal = new DFSGraphTraversal();
  graphTraversal->setComponentVisitor(new MeshDrawerCoreVisitor());
}

MeshDrawer::~MeshDrawer()
{
  GVManager::removeInstanceOf<MeshDrawerGlobalVariables>();
  
  delete graphTraversal;
}

void MeshDrawer::changeDrawPoints()
{
  drawPoints = !drawPoints;
}

void MeshDrawer::changeDrawEdges()
{
  drawEdges = !drawEdges;
}

void MeshDrawer::changeDrawFaces()
{
  drawFaces = !drawFaces;
}

void MeshDrawer::setSelectedType(SelectedType selectedType)
{
  this->selectedType = selectedType;
}

void MeshDrawer::setSelectedMesh(int selectedMesh)
{
  this->selectedMesh = selectedMesh;
}

void MeshDrawer::setSelectedId(int selectedId)
{
  this->selectedId = selectedId;
}

void MeshDrawer::drawMesh(X3DNode *N, bool selection)
{
  globalVariables->setSelectionMode(selection);
  
  if (selection || drawPoints)
  {
    globalVariables->setPrimitiveType(VERTICES);
    globalVariables->initTraversal();
    graphTraversal->traverse(N);
    globalVariables->finishTraversal();
  }  
  if (selection || drawEdges)
  {
    globalVariables->setPrimitiveType(EDGES);
    globalVariables->initTraversal();
    graphTraversal->traverse(N);
    globalVariables->finishTraversal();
  }  
  if (selection || drawFaces)
  {
    globalVariables->setPrimitiveType(FACES);  
    globalVariables->initTraversal();
    graphTraversal->traverse(N);
    globalVariables->finishTraversal();
  }  
}

void MeshDrawer::drawSelected()
{
  if (selectedType == NOTHING)
    return;

  glPushMatrix();
  glMultMatrixf(globalVariables->getMatrix(selectedMesh).toFloat16());

  const Mesh *mesh = globalVariables->getMesh(selectedMesh);
  bool normalPerVertex = mesh->data().getNormalPerVertex();

  glEnable(GL_POLYGON_OFFSET_FILL);
#ifdef GL_POLYGON_OFFSET_LINES
  glEnable(GL_POLYGON_OFFSET_LINES);
#endif
  
  glPointSize(15.0);
  glLineWidth(8.0);
  
  switch (selectedType)
  {
    case NOTHING :
      break;
    case VERTICES :
    {
      glPointSize(25.0);
      Mesh::Vertex* vertex = (mesh->getVertices())[selectedId];
      const Mesh::MFace& faces = vertex->getFaces();

      // Normal
      Mesh::Face* firstFace = (*(faces.begin()));
      SFVec3f normal;
      if (normalPerVertex)
        normal = vertex->data().getNormalOfFace(firstFace);
      else
        normal = firstFace->data().getNormal();
      glNormal3fv(normal.f_data());

      glColor3f(1.0, 1.0, 0.0);
      drawVertex(vertex);

      glColor3f(0.0, 0.8, 0.0);
      const Mesh::MEdge& edges = vertex->getEdges();
      for (Mesh::MEdge::const_iterator e = edges.begin(), end = edges.end() ; e != end; ++e)
        drawEdge(*e);

      glColor3f(0.7, 0.9, 0.7);
      for (Mesh::MFace::const_iterator f = faces.begin(), end = faces.end() ; f != end; ++f)
        drawFace(*f);
      break;
    }
    case EDGES :
    {
      glLineWidth(15.0);
      Mesh::Edge* edge = (mesh->getEdges())[selectedId];

      // Normal
      Mesh::Face* firstFace = (*(edge->getLeftFaces().begin()));
      if (!firstFace)
        firstFace = (*(edge->getRightFaces().begin()));
      SFVec3f normal;
      if (normalPerVertex)
        normal = edge->getFromVertex()->data().getNormalOfFace(firstFace);
      else
        normal = firstFace->data().getNormal();
      glNormal3fv(normal.f_data());

      glColor3f(1.0, 1.0, 0.0);
      drawEdge(edge);

      glColor3f(0.0, 0.8, 0.0);
      drawVertex(edge->getFromVertex());
      drawVertex(edge->getToVertex());

      glColor3f(0.9, 0.7, 0.7);
      const Mesh::MFace& lfaces = edge->getLeftFaces();
      for (Mesh::MFace::const_iterator f = lfaces.begin(), end = lfaces.end() ; f != end; ++f)
        drawFace(*f);

      glColor3f(0.7, 0.9, 0.7);
      const Mesh::MFace& rfaces = edge->getRightFaces();
      for (Mesh::MFace::const_iterator f = rfaces.begin(), end = rfaces.end() ; f != end; ++f)
        drawFace(*f);

      break;
    }
    case FACES :
    {
      Mesh::Face* face = (mesh->getFaces())[selectedId];
      const Mesh::MEdge& edges = face->getEdges();

      // Normal
      Mesh::Edge* firstEdge = (*(edges.begin()));
      SFVec3f normal;
      if (normalPerVertex)
        normal = firstEdge->getFromVertex()->data().getNormalOfFace(face);
      else
        normal = face->data().getNormal();
      glNormal3fv(normal.f_data());

      glColor3f(1.0, 1.0, 0.0);
      drawFace(face);

      for (Mesh::MEdge::const_iterator e = edges.begin(), end = edges.end() ; e != end; ++e)
      {
        glColor3f(0.0, 0.8, 0.0);
        drawEdge(*e);
        glColor3f(0.7, 0.7, 0.7);
        drawVertex((*e)->getFromVertex());
      }

      break;
    }
  }
  glDisable(GL_POLYGON_OFFSET_FILL);
#ifdef GL_POLYGON_OFFSET_LINES
  glDisable(GL_POLYGON_OFFSET_LINES);
#endif

  glPopMatrix();
}

void MeshDrawer::drawVertex(const Mesh::Vertex *vertex)
{
  glBegin(GL_POINTS);
  SFPoint3f point = vertex->data().getPoint();
  glVertex3fv(point.f_data());
  glEnd();
}

void MeshDrawer::drawEdge(Mesh::Edge *edge)
{
  glPolygonOffset(-100, -100);
  glDisable(GL_LIGHTING);
  glBegin(GL_LINES);
  SFPoint3f point = edge->getFromVertex()->data().getPoint();
  glVertex3fv(point.f_data());
  point = edge->getToVertex()->data().getPoint();
  glVertex3fv(point.f_data());
  glEnd();
  glEnable(GL_LIGHTING);
  glPolygonOffset(-2, -2);
}

void MeshDrawer::drawFace(const Mesh::Face *face)
{
  const Mesh::MEdge& edges = face->getEdges();
  glBegin(GL_POLYGON);
  for (Mesh::MEdge::const_iterator e = edges.begin(), end = edges.end() ; e != end; ++e)
  {
    SFPoint3f point = (*e)->getFromVertex()->data().getPoint();
    glVertex3fv(point.f_data());
  }
  glEnd();
}

}
}

