/// This file is part of the X3DToolKit library
/// Copyright (C) 2002-2003  Yannick Le Goc (legoc@imag.fr)

/// http://www-imagis.imag.fr/Membres/Yannick.Legoc/X3D/


/// X3DToolKit is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.

/// X3DToolKit is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.

/// You should have received a copy of the GNU General Public License
/// along with X3DToolKit; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
///////////////////////////////////////////////////////////////////////////////
//                            X3DComponentVisitor.h                           //
////////////////////////////////////////////////////////////////////////////////

#ifndef X3DCOMPONENTVISITOR_H
#define X3DCOMPONENTVISITOR_H

#include "X3DComponent.h"
#include "GVManager.h"
#include "EnterFunction.h"
#include "WalkOnFunction.h"
#include "LeaveFunction.h"
#include "X3DAbstractNode.h"
#include "SceneGraph.h"

namespace X3DTK {

/// Abstract class defining the concept of component visitor.

class X3DComponentVisitor : public X3DComponent
{
public:
  /// Constructor.
  X3DComponentVisitor();
  /// Destructor.
  virtual ~X3DComponentVisitor() = 0;
  
  /// Gets the enter function of name.
  EnterFunction *getEnterFunctionOf(const Type *type) const;
  /// Gets the WalkOn function of name.
  WalkOnFunction *getWalkOnFunctionOf(const Type *type) const;
  /// Gets the leave function of name.
  LeaveFunction *getLeaveFunctionOf(const Type *type) const;
  
protected:    
  /// Define a new enter function.
  template<class C, class T> 
  void defineNewEnterFunction(void (C::*ptrF)(T *) const)
  {
    Type::beginFunctionDefinition();
    
    EnterFunction *EF = new EnterFunction(reinterpret_cast<ptrToEnterFunction>(ptrF), this);
    //finding if an occurence of T
    T *t = new T();
    SFString name = t->getTypeName();
    Component *TComponent = t->getType()->getComponent();
    //checking if the node belongs to the component.
    if (TComponent != component)
    {
      SFString sg1, sg2;
      if (component->getSceneGraphName() != "")
        sg1 = " of the " + component->getSceneGraphName() + " scene graph";
      if (t->getType()->getSceneGraphName() != "")
        sg1 = " of the " + t->getType()->getSceneGraphName() + " scene graph";  

      std::cerr << "warning: defineNewEnterFunction for " << name << " which doesn't belong to the " << component->getName() << " component" << sg1 << "," << std::endl;
      std::cerr << "         but to the " << t->getType()->getComponentName() << " component" << sg2 << std::endl;
    }
    else
      _enterDict.insert(std::pair<SFString, EnterFunction *>(t->getType()->getEncodedName(), EF));

    delete t;
    Type::endFunctionDefinition();
  };
  
  /// Define a new walk On function.
  template<class C, class T> 
  void defineNewWalkOnFunction(bool (C::*ptrF)(T *, SFAbstractNode) const)
  {
    Type::beginFunctionDefinition();
    
    WalkOnFunction *WF = new WalkOnFunction(reinterpret_cast<ptrToWalkOnFunction>(ptrF), this);
    //finding if an occurence of T
    T *t = new T();
    SFString name = t->getTypeName();
    Component *TComponent = t->getType()->getComponent();
    //checking if the node belongs to the component.
    if (TComponent != component)
    {
      SFString sg1, sg2;
      if (component->getSceneGraphName() != "")
        sg1 = " of the " + component->getSceneGraphName() + " scene graph";
      if (t->getType()->getSceneGraphName() != "")
        sg1 = " of the " + t->getType()->getSceneGraphName() + " scene graph";  

      std::cerr << "warning: defineNewWalkOnFunction for " << name << " which doesn't belong to the " << component->getName() << " component" << sg1 << "," << std::endl;
      std::cerr << "         but to the " << t->getType()->getComponentName() << " component" << sg2 << std::endl;    
    }
    else
      _walkonDict.insert(std::pair<SFString, WalkOnFunction *>(t->getType()->getEncodedName(), WF));
      
    delete t;
    Type::endFunctionDefinition();
  };
  
  /// Define a new leave function.
  template<class C, class T> 
  void defineNewLeaveFunction(void (C::*ptrF)(T *) const)
  {
    Type::beginFunctionDefinition();
    
    LeaveFunction *LF = new LeaveFunction(reinterpret_cast<ptrToLeaveFunction>(ptrF), this);
    //finding if an occurence of T
    T *t = new T();
    SFString name = t->getTypeName();
    Component *TComponent = t->getType()->getComponent();
    //checking if the node belongs to the component.
    if (TComponent != component)
    {
      SFString sg1, sg2;
      if (component->getSceneGraphName() != "")
        sg1 = " of the " + component->getSceneGraphName() + " scene graph";
      if (t->getType()->getSceneGraphName() != "")
        sg1 = " of the " + t->getType()->getSceneGraphName() + " scene graph";  

      std::cerr << "warning: defineNewLeaveFunction for " << name << " which doesn't belong to the " << component->getName() << " component" << sg1 << "," << std::endl;
      std::cerr << "         but to the " << t->getType()->getComponentName() << " component" << sg2 << std::endl;    
    }
    else
      _leaveDict.insert(std::pair<SFString, LeaveFunction *>(t->getType()->getEncodedName(), LF));
      
    delete t;    
    Type::endFunctionDefinition();
  };
  
private:
  EnterDict _enterDict;  
  WalkOnDict _walkonDict;  
  LeaveDict _leaveDict;  
};

}

#endif
