#include "GL_IndexedFaceSet.h"
#include "IndexedFaceSet.h"
#include "Normal.h"
#include "Coordinate.h"
#include "TextureCoordinate.h"
#include "Color.h"
#include "ColorRGBA.h"

#include <list>

using namespace std;

namespace X3DTK {
namespace GL {

typedef struct
{
  SFInt32 face;
  unsigned short position;
} faceposition;

typedef struct
{
  list<SFInt32> faceList;
  SFVec3f normal;
} partnormal;

IndexedFaceSet::IndexedFaceSet() 
: X3DComposedGeometryNode()
{
  defineTypeName("IndexedFaceSet", "Geometry3D");
}
           
IndexedFaceSet::IndexedFaceSet(const IndexedFaceSet &I)
: X3DComposedGeometryNode(I), _N3F_V3F_vertexArray(I._N3F_V3F_vertexArray), _C4F_N3F_V3F_vertexArray(I._C4F_N3F_V3F_vertexArray), _T2F_N3F_V3F_vertexArray(I._T2F_N3F_V3F_vertexArray), _T2F_C4F_N3F_V3F_vertexArray(I._T2F_C4F_N3F_V3F_vertexArray), _indexArray(I._indexArray)
{
}

SFNode IndexedFaceSet::clone() const
{
  return new IndexedFaceSet(*this);
}

IndexedFaceSet::~IndexedFaceSet()
{
}

void IndexedFaceSet::setN3F_V3F_vertexArray(const std::vector<N3F_V3F> &N3F_V3F_vertexArray)
{
  _N3F_V3F_vertexArray = N3F_V3F_vertexArray;
}

void IndexedFaceSet::setC4F_N3F_V3F_vertexArray(const std::vector<C4F_N3F_V3F> &C4F_N3F_V3F_vertexArray)
{
  _C4F_N3F_V3F_vertexArray = C4F_N3F_V3F_vertexArray;
}

void IndexedFaceSet::setT2F_N3F_V3F_vertexArray(const std::vector<T2F_N3F_V3F> &T2F_N3F_V3F_vertexArray)
{
  _T2F_N3F_V3F_vertexArray = T2F_N3F_V3F_vertexArray;
}

void IndexedFaceSet::setT2F_C4F_N3F_V3F_vertexArray(const std::vector<T2F_C4F_N3F_V3F> &T2F_C4F_N3F_V3F_vertexArray)
{
  _T2F_C4F_N3F_V3F_vertexArray = T2F_C4F_N3F_V3F_vertexArray;
}

void IndexedFaceSet::setIndexArray(const MFInt32 &indexArray)
{
  _indexArray = indexArray;
}

void IndexedFaceSet::computeNormals()
{
  switch (getVertexFormat())
  {
    case GL_N3F_V3F:
      computeNoColorNoTexCoordNormals();
      break;
    case GL_C4F_N3F_V3F:
      computeColorNoTexCoordNormals();
      break;
    case GL_T2F_N3F_V3F:
      computeNoColorTexCoordNormals();
      break; 
    case GL_T2F_C4F_N3F_V3F:
      computeColorTexCoordNormals();
      break;
  }   
}

void IndexedFaceSet::update()
{
  if (x3dReference == 0)
    return;
    
  X3D::IndexedFaceSet *I = static_cast<X3D::IndexedFaceSet *>(x3dReference);  

  //for the moment if the node color has changed, all the attributes are computed
  //emptying the arrays
  emptyArrays();

  X3D::Coordinate *CoordNode = static_cast<X3D::Coordinate *>(I->getCoord());
  if (CoordNode == 0)
    return;

  const MFVec3f &vertexArray = CoordNode->getPoint();
  const MFInt32 &coordIndex = I->getCoordIndex();
  
  //IndexedFaceSet attributes  
  if (coordIndex.empty())
    return;
 
  setSolid(I->getSolid());
  setCcw(I->getCcw());
  setNormalPerVertex(I->getNormalPerVertex());

  computeIndexArray(coordIndex);

  MFVec3f array;
  MFVec3f &normalArray = array;

  X3D::Normal *NormalNode = dynamic_cast<X3D::Normal *>(I->getNormal());
    
  if (NormalNode != 0)
    normalArray = NormalNode->getVector();
  //in the other case, normalArray remains empty, that is the test to know whether the 
  //normals have to be assigned or not.

  //computeArrays(vertexArray, coordIndex, normalArray); 

  if ((I->getColor() != 0) && (I->getTexCoord() != 0))
  {
    setColor(true);
    setTexCoord(true);
    setVertexFormat(GL_T2F_C4F_N3F_V3F);
    computeColorTexCoord(vertexArray, normalArray, coordIndex);
  }
  if ((I->getColor() != 0) && (I->getTexCoord() == 0))
  {
    setColor(true);
    setTexCoord(false);
    setVertexFormat(GL_C4F_N3F_V3F);
    computeColorNoTexCoord(vertexArray, normalArray, coordIndex);
  }
  if ((I->getColor() == 0) && (I->getTexCoord() != 0))
  {
    setColor(false);
    setTexCoord(true);
    setVertexFormat(GL_T2F_N3F_V3F);
    computeNoColorTexCoord(vertexArray, normalArray, coordIndex);
  }
  if ((I->getColor() == 0) && (I->getTexCoord() == 0))
  {
    setColor(false);
    setTexCoord(false);
    setVertexFormat(GL_N3F_V3F);
    computeNoColorNoTexCoord(vertexArray, normalArray, coordIndex);
  }

  //Computes the normals.
  if (NormalNode == 0)
    computeNormals();
}

void IndexedFaceSet::render() const
{
  if (getSolid())
  {
    glEnable(GL_CULL_FACE);
    if (getCcw())
      glFrontFace(GL_CCW);
    else
      glFrontFace(GL_CW);
    glCullFace(GL_BACK);   
  }
  
  if ((getColor()) && (getTexCoord()))
  {
    glEnable(GL_COLOR_MATERIAL);
    if (!getVerticesDuplicated())
    {
      glInterleavedArrays(GL_T2F_C4F_N3F_V3F, 0, &_T2F_C4F_N3F_V3F_vertexArray.front());
      glDrawElements(GL_TRIANGLES, _indexArray.size(), GL_UNSIGNED_INT, &_indexArray.front());
    }
    else
    {
      glInterleavedArrays(GL_T2F_C4F_N3F_V3F, 0, &_T2F_C4F_N3F_V3F_vertexArray.front());
      glDrawArrays(GL_TRIANGLES, 0, _T2F_C4F_N3F_V3F_vertexArray.size());
    }
    glDisable(GL_COLOR_MATERIAL);
  }  
    
  if ((getColor()) && (!getTexCoord()))
  {
    glEnable(GL_COLOR_MATERIAL);
    if (!getVerticesDuplicated())
    {
      glInterleavedArrays(GL_C4F_N3F_V3F, 0, &_C4F_N3F_V3F_vertexArray.front());
      glDrawElements(GL_TRIANGLES, _indexArray.size(), GL_UNSIGNED_INT, &_indexArray.front());
    }
    else
    {
      glInterleavedArrays(GL_C4F_N3F_V3F, 0, &_C4F_N3F_V3F_vertexArray.front());
      glDrawArrays(GL_TRIANGLES, 0, _C4F_N3F_V3F_vertexArray.size());
    }
    glDisable(GL_COLOR_MATERIAL);
  }  
    
  if ((!getColor()) && (getTexCoord()))
  {
    glEnable(GL_TEXTURE_2D);
    if (!getVerticesDuplicated())
    {
      glInterleavedArrays(GL_T2F_N3F_V3F, 0, &_T2F_N3F_V3F_vertexArray.front());
      glDrawElements(GL_TRIANGLES, _indexArray.size(), GL_UNSIGNED_INT, &_indexArray.front());
    }
    else
    {
      glInterleavedArrays(GL_T2F_N3F_V3F, 0, &_T2F_N3F_V3F_vertexArray.front());
      glDrawArrays(GL_TRIANGLES, 0, _T2F_N3F_V3F_vertexArray.size());
    }
    glDisable(GL_TEXTURE_2D);
  }  
  
  if ((!getColor()) && (!getTexCoord()))
  {
    if (!getVerticesDuplicated())
    {
      glInterleavedArrays(GL_N3F_V3F, 0, &_N3F_V3F_vertexArray.front());
      glDrawElements(GL_TRIANGLES, _indexArray.size(), GL_UNSIGNED_INT, &_indexArray.front());
    }
    else
    {
      glInterleavedArrays(GL_N3F_V3F, 0, &_N3F_V3F_vertexArray.front());
      glDrawArrays(GL_TRIANGLES, 0, _N3F_V3F_vertexArray.size());
    }
  }
  
  if (getSolid())
    glDisable(GL_CULL_FACE);  
}

void IndexedFaceSet::emptyArrays()
{
  _N3F_V3F_vertexArray.clear();
  _C4F_N3F_V3F_vertexArray.clear(); 
  _T2F_N3F_V3F_vertexArray.clear();
  _T2F_C4F_N3F_V3F_vertexArray.clear(); 
  _indexArray.clear();
}

void IndexedFaceSet::computeIndexArray(const MFInt32 &coordIndex)
{
  //filling the index array
  //iterate the coordIndex
  _indexArray.reserve(3*coordIndex.size()/4);
  MFInt32::const_iterator itCoordIndex = coordIndex.begin();
  while (itCoordIndex != coordIndex.end())    
  {
    unsigned int first, a, b;
    
    if (*itCoordIndex == -1) {++itCoordIndex;continue;}
    
    first = *itCoordIndex;
    ++itCoordIndex;
    
    if (*itCoordIndex == -1) {++itCoordIndex;continue;}
    
    a = *itCoordIndex;
    ++itCoordIndex;
    
    while (*itCoordIndex != -1)
    {
      _indexArray.push_back(first);
      _indexArray.push_back(a);
      b = *itCoordIndex;
      ++itCoordIndex;
      _indexArray.push_back(b);
      a = b;
    }  
    ++itCoordIndex;
  }
}

void IndexedFaceSet::computeNoColorNoTexCoord(const MFVec3f &vertexArray, const MFVec3f &normalArray, const MFInt32 &coordIndex)
{
  X3D::IndexedFaceSet *I = static_cast<X3D::IndexedFaceSet *>(x3dReference);

  SFBool normalPerVertex = I->getNormalPerVertex();
  SFBool normalsComputed = !normalArray.empty();
  
  _X3DtoGLIndex.resize(vertexArray.size());
  
  //using the GL_N3F_V3F
  //vertex array
  if ((normalPerVertex) 
   && (I->getNormalIndex().empty()))
  {
    setVerticesDuplicated(false);
    //using the GLPOLYGON 
    //filling the vertex array
    MFVec3f::const_iterator itNormal = normalArray.begin();
       
    _N3F_V3F_vertexArray.resize(vertexArray.size());
    vector<N3F_V3F>::iterator itElement = _N3F_V3F_vertexArray.begin();
      
    for (MFVec3f::const_iterator itCoord = vertexArray.begin(); itCoord != vertexArray.end(); ++itCoord)
    {
      if (normalsComputed)
      {
        if (itNormal == normalArray.end()) {emptyArrays();return;}
        (*itElement).normal = *itNormal; 
        ++itNormal;
      }
      
      (*itElement).vertex = *itCoord;
      ++itElement;
    }
    
    unsigned int k = 0;
    for (vector<MFInt32>::iterator itIndex = _X3DtoGLIndex.begin(); itIndex != _X3DtoGLIndex.end(); ++itIndex)
    {
      (*itIndex).push_back(k);
      ++k;
    }
  }
  else 
  {
    setVerticesDuplicated(true);
    //we cannot use the index array because an attribute is not constant per
    //instantiation of vertex
    //using the GLPOLYGON 
    //filling the arrays
    _N3F_V3F_vertexArray.reserve(3*coordIndex.size()/4);
    
    const MFInt32 &normalIndex = (I->getNormalIndex().empty() ? coordIndex : I->getNormalIndex());
        
    MFInt32::const_iterator itCoordIndex = coordIndex.begin();
    MFInt32::const_iterator itNormalIndex = normalIndex.begin();
    
    while (itCoordIndex != coordIndex.end())    
    {
      N3F_V3F first, a, b;
      SFInt32 index_first, index_a, index_b;
      
      if (*itCoordIndex == -1) {++itCoordIndex;++itNormalIndex;continue;}
      
      ///////////////////////////computing first//////////////////////////////////////
      if (itNormalIndex == normalIndex.end()) {emptyArrays();return;}
      if (normalsComputed)
      {
        if (static_cast<unsigned int>(*itNormalIndex) >= normalArray.size()) {emptyArrays();return;}
        first.normal = normalArray[*itNormalIndex];
        if (normalPerVertex)
          ++itNormalIndex;      
      }
      
      if (static_cast<unsigned int>(*itCoordIndex) >= vertexArray.size()) {emptyArrays();return;}
      first.vertex = vertexArray[*itCoordIndex];
      index_first = *itCoordIndex;
      ++itCoordIndex;
            
      ////////////////////////////////////////////////////////////////////////////////     
            
      if (*itCoordIndex == -1) {++itCoordIndex;++itNormalIndex;continue;}
      
      ///////////////////////////computing a//////////////////////////////////////////
      if (itNormalIndex == normalIndex.end()) {emptyArrays();return;}
      if (normalsComputed)
      {
        if (static_cast<unsigned int>(*itNormalIndex) >= normalArray.size()) {emptyArrays();return;}
        a.normal = normalArray[*itNormalIndex];
        if (normalPerVertex)
          ++itNormalIndex;      
      }
            
      if (static_cast<unsigned int>(*itCoordIndex) >= vertexArray.size()) {emptyArrays();return;}
      a.vertex = vertexArray[*itCoordIndex];
      index_a = *itCoordIndex;
      ++itCoordIndex;
            
      ////////////////////////////////////////////////////////////////////////////////
      
      while (*itCoordIndex != -1)
      {
        ///////////////////////////computing b//////////////////////////////////////////
        if (itNormalIndex == normalIndex.end()) {emptyArrays();return;}
        if (normalsComputed)
        {
          if (static_cast<unsigned int>(*itNormalIndex) >= normalArray.size()) {emptyArrays();return;}
          b.normal = normalArray[*itNormalIndex];
          if (normalPerVertex)
            ++itNormalIndex;      
        }
            
        if (static_cast<unsigned int>(*itCoordIndex) >= vertexArray.size()) {emptyArrays();return;}
        b.vertex = vertexArray[*itCoordIndex];
        index_b = *itCoordIndex;
        ++itCoordIndex;
              
        ////////////////////////////////////////////////////////////////////////////////
                
        //pushing back first, a, b     
        _N3F_V3F_vertexArray.push_back(first);
        _X3DtoGLIndex[index_first].push_back(_N3F_V3F_vertexArray.size() - 1);
        
        _N3F_V3F_vertexArray.push_back(a);
        _X3DtoGLIndex[index_a].push_back(_N3F_V3F_vertexArray.size() - 1);
        
        _N3F_V3F_vertexArray.push_back(b);
        _X3DtoGLIndex[index_b].push_back(_N3F_V3F_vertexArray.size() - 1);
        
        //assigning b to a
        a.normal = b.normal;
        a.vertex = b.vertex;
        index_a = index_b;
      }
      ++itNormalIndex;
      ++itCoordIndex;
    }
  }
}
  
void IndexedFaceSet::computeColorNoTexCoord(const MFVec3f &vertexArray, const MFVec3f &normalArray, const MFInt32 &coordIndex)
{
  X3D::IndexedFaceSet *I = static_cast<X3D::IndexedFaceSet *>(x3dReference);

  const MFColorRGBA &colorRGBA = (I->getColor()->getTypeName() == "ColorRGBA" ? dynamic_cast<X3D::ColorRGBA *>(I->getColor())->getColor() : dynamic_cast<X3D::Color *>(I->getColor())->getColor());
  
  SFBool normalPerVertex = I->getNormalPerVertex();
  SFBool colorPerVertex = I->getColorPerVertex();
  SFBool normalsComputed = !normalArray.empty();
  
  _X3DtoGLIndex.resize(vertexArray.size());
  
  //using the GL_C4F_N3F_V3F
  //vertex array
  if ((normalPerVertex)
   && (colorPerVertex) 
   && (I->getColorIndex().empty()) 
   && (I->getNormalIndex().empty()))
  {
    setVerticesDuplicated(false);
    //using the GLPOLYGON 
    //filling the vertex array
    MFColorRGBA::const_iterator itColor = colorRGBA.begin();
    MFVec3f::const_iterator itNormal = normalArray.begin();
    
    _C4F_N3F_V3F_vertexArray.resize(vertexArray.size());
    vector<C4F_N3F_V3F>::iterator itElement = _C4F_N3F_V3F_vertexArray.begin();
        
    for (MFVec3f::const_iterator itCoord = vertexArray.begin(); itCoord != vertexArray.end(); ++itCoord)
    {
      if (itColor == colorRGBA.end()) {emptyArrays();return;}
      (*itElement).color = *itColor;
      ++itColor;

      if (normalsComputed)
      {
        if (itNormal == normalArray.end()) {emptyArrays();return;}
        (*itElement).normal = *itNormal; 
        ++itNormal;
      }
            
      (*itElement).vertex = *itCoord;
      ++itElement;
    }
    
    unsigned int k = 0;
    for (vector<MFInt32>::iterator itIndex = _X3DtoGLIndex.begin(); itIndex != _X3DtoGLIndex.end(); ++itIndex)
    {
      (*itIndex).push_back(k);
      ++k;
    }
  }
  else //normalPerVertex or colorPertex is FALSE or ...
  {
    setVerticesDuplicated(true);
    //we cannot use the index array because an attribute is not constant per
    //instantiation of vertex
    //using the GLPOLYGON 
    //filling the vertex array
    //filling the arrays
    _C4F_N3F_V3F_vertexArray.reserve(3*coordIndex.size()/4);
    
    const MFInt32 &colorIndex = (I->getColorIndex().empty() ? coordIndex : I->getColorIndex());
    const MFInt32 &normalIndex = (I->getNormalIndex().empty() ? coordIndex : I->getNormalIndex());
        
    MFInt32::const_iterator itCoordIndex = coordIndex.begin();
    MFInt32::const_iterator itColorIndex = colorIndex.begin();
    MFInt32::const_iterator itNormalIndex = normalIndex.begin();
    
    while (itCoordIndex != coordIndex.end())    
    {
      C4F_N3F_V3F first, a, b;
      SFInt32 index_first, index_a, index_b;
      
      if (*itCoordIndex == -1) {++itCoordIndex;++itColorIndex;++itNormalIndex;continue;}
      
      ///////////////////////////computing first//////////////////////////////////////
      if ((itColorIndex == colorIndex.end()) || (static_cast<unsigned int>(*itColorIndex) >= colorRGBA.size())) {emptyArrays();return;}
      first.color = colorRGBA[*itColorIndex];
      if (colorPerVertex)
        ++itColorIndex;           
      
      if (itNormalIndex == normalIndex.end()) {emptyArrays();return;}
      if (normalsComputed)
      {
        if (static_cast<unsigned int>(*itNormalIndex) >= normalArray.size()) {emptyArrays();return;}
        first.normal = normalArray[*itNormalIndex];
        if (normalPerVertex)
          ++itNormalIndex;      
      }
      
      if (static_cast<unsigned int>(*itCoordIndex) >= vertexArray.size()) {emptyArrays();return;}
      first.vertex = vertexArray[*itCoordIndex];
      index_first = *itCoordIndex;
      ++itCoordIndex;
            
      ////////////////////////////////////////////////////////////////////////////////     
      
      if (*itCoordIndex == -1) {++itCoordIndex;++itColorIndex;++itNormalIndex;continue;}
      
      ///////////////////////////computing a//////////////////////////////////////////
      if ((itColorIndex == colorIndex.end()) || (static_cast<unsigned int>(*itColorIndex) >= colorRGBA.size())) {emptyArrays();return;}
      a.color = colorRGBA[*itColorIndex];
      if (colorPerVertex)
        ++itColorIndex;           
      
      if (itNormalIndex == normalIndex.end()) {emptyArrays();return;}
      if (normalsComputed)
      {
        if (static_cast<unsigned int>(*itNormalIndex) >= normalArray.size()) {emptyArrays();return;}
        a.normal = normalArray[*itNormalIndex];
        if (normalPerVertex)
          ++itNormalIndex;      
      }
      
      if (static_cast<unsigned int>(*itCoordIndex) >= vertexArray.size()) {emptyArrays();return;}
      a.vertex = vertexArray[*itCoordIndex];
      index_a = *itCoordIndex;
      ++itCoordIndex;
            
      ////////////////////////////////////////////////////////////////////////////////     
      
      while (*itCoordIndex != -1)
      {
        ///////////////////////////computing b//////////////////////////////////////
        if ((itColorIndex == colorIndex.end()) || (static_cast<unsigned int>(*itColorIndex) >= colorRGBA.size())) {emptyArrays();return;}
        b.color = colorRGBA[*itColorIndex];
        if (colorPerVertex)
          ++itColorIndex;           
      
        if (itNormalIndex == normalIndex.end()) {emptyArrays();return;}
        if (normalsComputed)
        {
          if (static_cast<unsigned int>(*itNormalIndex) >= normalArray.size()) {emptyArrays();return;}
          b.normal = normalArray[*itNormalIndex];
          if (normalPerVertex)
            ++itNormalIndex;      
        }
      
        if (static_cast<unsigned int>(*itCoordIndex) >= vertexArray.size()) {emptyArrays();return;}
        b.vertex = vertexArray[*itCoordIndex];
        index_b = *itCoordIndex;
        ++itCoordIndex;   
        
        ////////////////////////////////////////////////////////////////////////////////     
              
        //pushing back first, a     
        _C4F_N3F_V3F_vertexArray.push_back(first);
        _X3DtoGLIndex[index_first].push_back(_C4F_N3F_V3F_vertexArray.size() - 1);
        
        _C4F_N3F_V3F_vertexArray.push_back(a);
        _X3DtoGLIndex[index_a].push_back(_C4F_N3F_V3F_vertexArray.size() - 1);
        
        _C4F_N3F_V3F_vertexArray.push_back(b);
        _X3DtoGLIndex[index_b].push_back(_C4F_N3F_V3F_vertexArray.size() - 1);
        
        //assigning b to a
        a.color = b.color;
        a.normal = b.normal;
        a.vertex = b.vertex;
        index_a = index_b;
      }
      ++itColorIndex;  
      ++itNormalIndex;
      ++itCoordIndex;
    }
  }
}

void IndexedFaceSet::computeNoColorTexCoord(const MFVec3f &vertexArray, const MFVec3f &normalArray, const MFInt32 &coordIndex)
{
  X3D::IndexedFaceSet *I = static_cast<X3D::IndexedFaceSet *>(x3dReference);

  SFBool normalPerVertex = I->getNormalPerVertex();
  SFBool normalsComputed = !normalArray.empty();
  
  _X3DtoGLIndex.resize(vertexArray.size());
  
  X3D::TextureCoordinate *TexCoordNode = dynamic_cast<X3D::TextureCoordinate *>(I->getTexCoord());
  if (TexCoordNode == 0)
    return;
  
  const MFVec2f &texCoordArray = TexCoordNode->getPoint();
 
  //using the GL_T2F_N3F_V3F
  //vertex array
  if ((normalPerVertex) 
   && (I->getNormalIndex().empty())
   && (I->getTexCoordIndex().empty()))
  {
    setVerticesDuplicated(false);
    //using the GLPOLYGON 
    //filling the vertex array
    MFVec2f::const_iterator itTexCoord = texCoordArray.begin();
    MFVec3f::const_iterator itNormal = normalArray.begin();
        
    _T2F_N3F_V3F_vertexArray.resize(vertexArray.size());
    vector<T2F_N3F_V3F>::iterator itElement = _T2F_N3F_V3F_vertexArray.begin();
      
    for (MFVec3f::const_iterator itCoord = vertexArray.begin(); itCoord != vertexArray.end(); ++itCoord)
    {
      if (itTexCoord == texCoordArray.end()) {emptyArrays();return;}
      (*itElement).texCoord = *itTexCoord;
      ++itTexCoord;
      
      if (normalsComputed)
      {
        if (itNormal == normalArray.end()) {emptyArrays();return;}
        (*itElement).normal = *itNormal; 
        ++itNormal;
      }
      
      (*itElement).vertex = *itCoord;
      ++itElement;
    }
  }
  else //normalPerVertex is FALSE or ...
  {
    setVerticesDuplicated(true);
    //we cannot use the index array because an attribute is not constant per
    //instantiation of vertex
    //using the GLPOLYGON 
    //filling the arrays
    _T2F_N3F_V3F_vertexArray.reserve(3*coordIndex.size()/4);
    
    const MFInt32 &texCoordIndex = (I->getTexCoordIndex().empty() ? coordIndex : I->getTexCoordIndex());
    const MFInt32 &normalIndex = (I->getNormalIndex().empty() ? coordIndex : I->getNormalIndex());
            
    MFInt32::const_iterator itCoordIndex = coordIndex.begin();
    MFInt32::const_iterator itNormalIndex = normalIndex.begin();
    MFInt32::const_iterator itTexCoordIndex = texCoordIndex.begin();
    
    while (itCoordIndex != coordIndex.end())    
    {
      T2F_N3F_V3F first, a, b;
      SFInt32 index_first, index_a, index_b;
            
      if (*itCoordIndex == -1) {++itCoordIndex;++itTexCoordIndex;++itNormalIndex;continue;}
      
      ///////////////////////////computing first//////////////////////////////////////
      if ((itTexCoordIndex == texCoordIndex.end()) || (static_cast<unsigned int>(*itTexCoordIndex) >= texCoordArray.size())) {emptyArrays();return;}
      first.texCoord = texCoordArray[*itTexCoordIndex];
      ++itTexCoordIndex;
            
      if (itNormalIndex == normalIndex.end()) {emptyArrays();return;}
      if (normalsComputed)
      {
        if (static_cast<unsigned int>(*itNormalIndex) >= normalArray.size()) {emptyArrays();return;}
        first.normal = normalArray[*itNormalIndex];
        if (normalPerVertex)
          ++itNormalIndex;      
      }
      
      if (static_cast<unsigned int>(*itCoordIndex) >= vertexArray.size()) {emptyArrays();return;}
      first.vertex = vertexArray[*itCoordIndex];
      index_first = *itCoordIndex;
      ++itCoordIndex;
            
      ////////////////////////////////////////////////////////////////////////////////     
      
      if (*itCoordIndex == -1) {++itCoordIndex;++itTexCoordIndex;++itNormalIndex;continue;}
      
      ///////////////////////////computing a//////////////////////////////////////
      if ((itTexCoordIndex == texCoordIndex.end()) || (static_cast<unsigned int>(*itTexCoordIndex) >= texCoordArray.size())) {emptyArrays();return;}
      a.texCoord = texCoordArray[*itTexCoordIndex];
      ++itTexCoordIndex;
            
      if (itNormalIndex == normalIndex.end()) {emptyArrays();return;}
      if (normalsComputed)
      {
        if (static_cast<unsigned int>(*itNormalIndex) >= normalArray.size()) {emptyArrays();return;}
        a.normal = normalArray[*itNormalIndex];
        if (normalPerVertex)
          ++itNormalIndex;      
      }
      
      if (static_cast<unsigned int>(*itCoordIndex) >= vertexArray.size()) {emptyArrays();return;}
      a.vertex = vertexArray[*itCoordIndex];
      index_a = *itCoordIndex;
      ++itCoordIndex;
            
      ////////////////////////////////////////////////////////////////////////////////     
      
      while (*itCoordIndex != -1)
      {
        ///////////////////////////computing b//////////////////////////////////////
        if ((itTexCoordIndex == texCoordIndex.end()) || (static_cast<unsigned int>(*itTexCoordIndex) >= texCoordArray.size())) {emptyArrays();return;}
        b.texCoord = texCoordArray[*itTexCoordIndex];
        ++itTexCoordIndex;
            
        if (itNormalIndex == normalIndex.end()) {emptyArrays();return;}
        if (normalsComputed)
        {
          if (static_cast<unsigned int>(*itNormalIndex) >= normalArray.size()) {emptyArrays();return;}
          b.normal = normalArray[*itNormalIndex];
          if (normalPerVertex)
            ++itNormalIndex;      
        }
      
        if (static_cast<unsigned int>(*itCoordIndex) >= vertexArray.size()) {emptyArrays();return;}
        b.vertex = vertexArray[*itCoordIndex];
        index_b = *itCoordIndex;
        ++itCoordIndex;
             
        ////////////////////////////////////////////////////////////////////////////////     
      
        //pushing back first, a     
        _T2F_N3F_V3F_vertexArray.push_back(first);
        _X3DtoGLIndex[index_first].push_back(_T2F_N3F_V3F_vertexArray.size() - 1);
        
        _T2F_N3F_V3F_vertexArray.push_back(a);
        _X3DtoGLIndex[index_a].push_back(_T2F_N3F_V3F_vertexArray.size() - 1);
        
        _T2F_N3F_V3F_vertexArray.push_back(b);
        _X3DtoGLIndex[index_b].push_back(_T2F_N3F_V3F_vertexArray.size() - 1);
        
        //assigning b to a
        a.texCoord = b.texCoord;
        a.normal = b.normal;
        a.vertex = b.vertex;
        index_a = index_b;
      }
      ++itTexCoordIndex;
      ++itNormalIndex;
      ++itCoordIndex;
    }
  }
}

void IndexedFaceSet::computeColorTexCoord(const MFVec3f &vertexArray, const MFVec3f &normalArray, const MFInt32 &coordIndex)
{
  //cout << "IndexedFaceSet: color, texCoord" << endl;
  X3D::IndexedFaceSet *I = static_cast<X3D::IndexedFaceSet *>(x3dReference);
  
  SFBool normalPerVertex = I->getNormalPerVertex();
  SFBool colorPerVertex = I->getColorPerVertex();
  SFBool normalsComputed = !normalArray.empty();
  
  _X3DtoGLIndex.resize(vertexArray.size());
  
  const MFColorRGBA &colorRGBA = (I->getColor()->getTypeName() == "ColorRGBA" ? dynamic_cast<X3D::ColorRGBA *>(I->getColor())->getColor() : dynamic_cast<X3D::Color *>(I->getColor())->getColor());
    
  X3D::TextureCoordinate *TexCoordNode = static_cast<X3D::TextureCoordinate *>(I->getTexCoord());
  if (TexCoordNode == 0)
    return;
  
  const MFVec2f texCoordArray = TexCoordNode->getPoint();
    
  //using the GL_T2F_C4F_N3F_V3F
  //vertex array
  if ((normalPerVertex)
   && (colorPerVertex) 
   && (I->getColorIndex().empty()) 
   && (I->getNormalIndex().empty())
   && (I->getTexCoordIndex().empty()))
  {
    setVerticesDuplicated(false);
    //using the GLPOLYGON 
    //filling the vertex array
    MFColorRGBA::const_iterator itColor = colorRGBA.begin();
    MFVec3f::const_iterator itNormal = normalArray.begin();
    MFVec2f::const_iterator itTexCoord = texCoordArray.begin();
    
    _T2F_C4F_N3F_V3F_vertexArray.resize(vertexArray.size());
        
    vector<T2F_C4F_N3F_V3F>::iterator itElement = _T2F_C4F_N3F_V3F_vertexArray.begin();
        
    for (MFVec3f::const_iterator itCoord = vertexArray.begin(); itCoord != vertexArray.end(); ++itCoord)
    {
      if (itNormal == normalArray.end()) {emptyArrays();return;}
      
      if (itTexCoord == texCoordArray.end()) {emptyArrays();return;}
      (*itElement).texCoord = *itTexCoord;
      ++itTexCoord;
      
      if (itColor == colorRGBA.end()) {emptyArrays();return;}
      (*itElement).color = *itColor;
      ++itColor;
      
      if (normalsComputed)
      {
        if (itNormal == normalArray.end()) {emptyArrays();return;}
        (*itElement).normal = *itNormal; 
        ++itNormal;
      }
      
      (*itElement).vertex = *itCoord;
      ++itElement;
    }
  }
  else //normalPerVertex or colorPertex is FALSE or ...
  {
    setVerticesDuplicated(true);
    //we cannot use the index array because an attribute is not constant per
    //instantiation of vertex
    //using the GLPOLYGON 
    //filling the vertex array
    //filling the arrays
    _T2F_C4F_N3F_V3F_vertexArray.reserve(3*coordIndex.size()/4);
    
    const MFInt32 &texCoordIndex = (I->getTexCoordIndex().empty() ? coordIndex : I->getTexCoordIndex());
    const MFInt32 &colorIndex = (I->getColorIndex().empty() ? coordIndex : I->getColorIndex());
    const MFInt32 &normalIndex = (I->getNormalIndex().empty() ? coordIndex : I->getNormalIndex());
            
    MFInt32::const_iterator itCoordIndex = coordIndex.begin();
    MFInt32::const_iterator itColorIndex = colorIndex.begin();
    MFInt32::const_iterator itNormalIndex = normalIndex.begin();
    MFInt32::const_iterator itTexCoordIndex = texCoordIndex.begin();
    
    while (itCoordIndex != coordIndex.end())    
    {
      T2F_C4F_N3F_V3F first, a, b;
      SFInt32 index_first, index_a, index_b;
      
      if (*itCoordIndex == -1) {++itCoordIndex;++itColorIndex;++itTexCoordIndex;++itNormalIndex;continue;}
      
      ///////////////////////////computing first//////////////////////////////////////
      if ((itTexCoordIndex == texCoordIndex.end()) || (static_cast<unsigned int>(*itTexCoordIndex) >= texCoordArray.size())) {emptyArrays();return;}
      first.texCoord = texCoordArray[*itTexCoordIndex];
      ++itTexCoordIndex;
            
      if ((itColorIndex == colorIndex.end()) || (static_cast<unsigned int>(*itColorIndex) >= colorRGBA.size())) {emptyArrays();return;}
      first.color = colorRGBA[*itColorIndex];
      if (colorPerVertex)
        ++itColorIndex;           
      
      if (itNormalIndex == normalIndex.end()) {emptyArrays();return;}
      if (normalsComputed)
      {
        if (static_cast<unsigned int>(*itNormalIndex) >= normalArray.size()) {emptyArrays();return;}
        first.normal = normalArray[*itNormalIndex];
        if (normalPerVertex)
          ++itNormalIndex;      
      }
      
      if (static_cast<unsigned int>(*itCoordIndex) >= vertexArray.size()) {emptyArrays();return;}
      first.vertex = vertexArray[*itCoordIndex];
      index_first = *itCoordIndex;
      ++itCoordIndex;
            
      ////////////////////////////////////////////////////////////////////////////////     
      
      if (*itCoordIndex == -1) {++itCoordIndex;++itColorIndex;++itTexCoordIndex;++itNormalIndex;continue;}
      
      ///////////////////////////computing a//////////////////////////////////////
      if ((itTexCoordIndex == texCoordIndex.end()) || (static_cast<unsigned int>(*itTexCoordIndex) >= texCoordArray.size())) {emptyArrays();return;}
      a.texCoord = texCoordArray[*itTexCoordIndex];
      ++itTexCoordIndex;
            
      if ((itColorIndex == colorIndex.end()) || (static_cast<unsigned int>(*itColorIndex) >= colorRGBA.size())) {emptyArrays();return;}
      a.color = colorRGBA[*itColorIndex];
      if (colorPerVertex)
        ++itColorIndex;           
      
      if (itNormalIndex == normalIndex.end()) {emptyArrays();return;}
      if (normalsComputed)
      {
        if (static_cast<unsigned int>(*itNormalIndex) >= normalArray.size()) {emptyArrays();return;}
        a.normal = normalArray[*itNormalIndex];
        if (normalPerVertex)
          ++itNormalIndex;      
      }
      
      if (static_cast<unsigned int>(*itCoordIndex) >= vertexArray.size()) {emptyArrays();return;}
      a.vertex = vertexArray[*itCoordIndex];
      index_a = *itCoordIndex;
      ++itCoordIndex;
            
      ////////////////////////////////////////////////////////////////////////////////     
      
      while (*itCoordIndex != -1)
      {
        ///////////////////////////computing b//////////////////////////////////////
        if ((itTexCoordIndex == texCoordIndex.end()) || (static_cast<unsigned int>(*itTexCoordIndex) >= texCoordArray.size())) {emptyArrays();return;}
        b.texCoord = texCoordArray[*itTexCoordIndex];
        ++itTexCoordIndex;
            
        if ((itColorIndex == colorIndex.end()) || (static_cast<unsigned int>(*itColorIndex) >= colorRGBA.size())) {emptyArrays();return;}
        b.color = colorRGBA[*itColorIndex];
        if (colorPerVertex)
          ++itColorIndex;           
      
        if (itNormalIndex == normalIndex.end()) {emptyArrays();return;}
        if (normalsComputed)
        {
          if (static_cast<unsigned int>(*itNormalIndex) >= normalArray.size()) {emptyArrays();return;}
          b.normal = normalArray[*itNormalIndex];
          if (normalPerVertex)
            ++itNormalIndex;      
        }
      
        if (static_cast<unsigned int>(*itCoordIndex) >= vertexArray.size()) {emptyArrays();return;}
        b.vertex = vertexArray[*itCoordIndex];
        index_b = *itCoordIndex;
        ++itCoordIndex;
            
        ////////////////////////////////////////////////////////////////////////////////     
            
        //pushing back first, a, b
        _T2F_C4F_N3F_V3F_vertexArray.push_back(first);
        _X3DtoGLIndex[index_first].push_back(_T2F_C4F_N3F_V3F_vertexArray.size() - 1);
        
        _T2F_C4F_N3F_V3F_vertexArray.push_back(a);
        _X3DtoGLIndex[index_a].push_back(_T2F_C4F_N3F_V3F_vertexArray.size() - 1);
        
        _T2F_C4F_N3F_V3F_vertexArray.push_back(b);
        _X3DtoGLIndex[index_b].push_back(_T2F_C4F_N3F_V3F_vertexArray.size() - 1);
        
        //assigning b to a
        a.texCoord = b.texCoord;
        a.color = b.color;
        a.normal = b.normal;
        a.vertex = b.vertex;
        index_a = index_b;
      }
      ++itTexCoordIndex;
      ++itColorIndex;  
      ++itNormalIndex;
      ++itCoordIndex;
    }
  }
}

void IndexedFaceSet::computeNoColorNoTexCoordNormals()
{
  MFVec3f normalFaceArray;
  
  if (getNormalPerVertex())
    normalFaceArray.reserve(_indexArray.size()/3);
  
  MFInt32::const_iterator itIndex = _indexArray.begin();
  
  // computing the normals per face.    
  while (itIndex != _indexArray.end())
  {
    SFVec3f A = _N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].vertex; 
    ++itIndex;
    
    SFVec3f B = _N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].vertex;
    ++itIndex;
    
    SFVec3f C = _N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].vertex;
    ++itIndex;
    
    SFVec3f AB = (B - A).normalize();
    SFVec3f AC = (C - A).normalize();
    SFVec3f N;

    if (getCcw())
      N = crossprod(AB, AC);
    else
      N = crossprod(AC, AB);  
    
    normalFaceArray.push_back(N);
  }
  
  MFVec3f::const_iterator itNormal = normalFaceArray.begin();
  if (getNormalPerVertex())
  {
    // adding normals per vertex.  
    itIndex = _indexArray.begin();
    while (itIndex != _indexArray.end())
    {
      _N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].normal += *itNormal;
      ++itIndex;
      _N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].normal += *itNormal;
      ++itIndex;
      _N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].normal += *itNormal;
      ++itIndex;
      ++itNormal;       
    }
    
    // copying and normalizing to identical values.
    for (vector<MFInt32>::const_iterator itIndexList = _X3DtoGLIndex.begin(); itIndexList != _X3DtoGLIndex.end(); ++itIndexList)
    {
      if (!(*itIndexList).empty())
      {
        _N3F_V3F_vertexArray[(*itIndexList).front()].normal.normalize();
        SFVec3f N = _N3F_V3F_vertexArray[(*itIndexList).front()].normal;

        MFInt32::const_iterator itBegin = (*itIndexList).begin();
        for (MFInt32::const_iterator itIndexDuplicated = ++itBegin; itIndexDuplicated != (*itIndexList).end(); ++itIndexDuplicated)
          _N3F_V3F_vertexArray[*itIndexDuplicated].normal = N;
      }  
    }
  }
  else
  {
    MFVec3f::const_iterator itNormal = normalFaceArray.begin();
    vector<N3F_V3F>::iterator itVertex = _N3F_V3F_vertexArray.begin();
    while (itVertex != _N3F_V3F_vertexArray.end())
    {
      (*itVertex).normal = *itNormal;
      ++itVertex;
      (*itVertex).normal = *itNormal;
      ++itVertex;
      (*itVertex).normal = *itNormal;
      ++itVertex;
      
      ++itNormal;
    }
  }
}

void IndexedFaceSet::computeColorNoTexCoordNormals()
{
  MFVec3f normalFaceArray;
  
  if (getNormalPerVertex())
    normalFaceArray.reserve(_indexArray.size()/3);
  
  MFInt32::const_iterator itIndex = _indexArray.begin();
  
  // computing the normals per face.    
  while (itIndex != _indexArray.end())
  {
    SFVec3f A = _C4F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].vertex; 
    ++itIndex;
    
    SFVec3f B = _C4F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].vertex;
    ++itIndex;
    
    SFVec3f C = _C4F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].vertex;
    ++itIndex;
    
    SFVec3f AB = (B - A).normalize();
    SFVec3f AC = (C - A).normalize();
    SFVec3f N;

    if (getCcw())
      N = crossprod(AB, AC);
    else
      N = crossprod(AC, AB);  
    
    normalFaceArray.push_back(N);
  }
  
  MFVec3f::const_iterator itNormal = normalFaceArray.begin();
  if (getNormalPerVertex())
  {
    // adding normals per vertex.  
    itIndex = _indexArray.begin();
    while (itIndex != _indexArray.end())
    {
      _C4F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].normal += *itNormal;
      ++itIndex;
      _C4F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].normal += *itNormal;
      ++itIndex;
      _C4F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].normal += *itNormal;
      ++itIndex;
      ++itNormal;       
    }
    
    // copying and normalizing to identical values.
    for (vector<MFInt32>::const_iterator itIndexList = _X3DtoGLIndex.begin(); itIndexList != _X3DtoGLIndex.end(); ++itIndexList)
    {
      if (!(*itIndexList).empty())
      {
        _C4F_N3F_V3F_vertexArray[(*itIndexList).front()].normal.normalize();
        SFVec3f N = _C4F_N3F_V3F_vertexArray[(*itIndexList).front()].normal;

        MFInt32::const_iterator itBegin = (*itIndexList).begin();
        for (MFInt32::const_iterator itIndexDuplicated = ++itBegin; itIndexDuplicated != (*itIndexList).end(); ++itIndexDuplicated)
          _C4F_N3F_V3F_vertexArray[*itIndexDuplicated].normal = N;
      }  
    }
  }
  else
  {
    MFVec3f::const_iterator itNormal = normalFaceArray.begin();
    vector<C4F_N3F_V3F>::iterator itVertex = _C4F_N3F_V3F_vertexArray.begin();
    while (itVertex != _C4F_N3F_V3F_vertexArray.end())
    {
      (*itVertex).normal = *itNormal;
      ++itVertex;
      (*itVertex).normal = *itNormal;
      ++itVertex;
      (*itVertex).normal = *itNormal;
      ++itVertex;
      
      ++itNormal;
    }
  }
}

void IndexedFaceSet::computeNoColorTexCoordNormals()
{
  MFVec3f normalFaceArray;
  
  if (getNormalPerVertex())
    normalFaceArray.reserve(_indexArray.size()/3);
  
  MFInt32::const_iterator itIndex = _indexArray.begin();
  
  // computing the normals per face.    
  while (itIndex != _indexArray.end())
  {
    SFVec3f A = _T2F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].vertex; 
    ++itIndex;
    
    SFVec3f B = _T2F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].vertex;
    ++itIndex;
    
    SFVec3f C = _T2F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].vertex;
    ++itIndex;
    
    SFVec3f AB = (B - A).normalize();
    SFVec3f AC = (C - A).normalize();
    SFVec3f N;

    if (getCcw())
      N = crossprod(AB, AC);
    else
      N = crossprod(AC, AB);  
    
    normalFaceArray.push_back(N);
  }
  
  MFVec3f::const_iterator itNormal = normalFaceArray.begin();
  if (getNormalPerVertex())
  {
    // adding normals per vertex.  
    itIndex = _indexArray.begin();
    while (itIndex != _indexArray.end())
    {
      _T2F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].normal += *itNormal;
      ++itIndex;
      _T2F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].normal += *itNormal;
      ++itIndex;
      _T2F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].normal += *itNormal;
      ++itIndex;
      ++itNormal;       
    }
    
    // copying and normalizing to identical values.
    for (vector<MFInt32>::const_iterator itIndexList = _X3DtoGLIndex.begin(); itIndexList != _X3DtoGLIndex.end(); ++itIndexList)
    {
      if (!(*itIndexList).empty())
      {
        _T2F_N3F_V3F_vertexArray[(*itIndexList).front()].normal.normalize();
        SFVec3f N = _T2F_N3F_V3F_vertexArray[(*itIndexList).front()].normal;

        MFInt32::const_iterator itBegin = (*itIndexList).begin();
        for (MFInt32::const_iterator itIndexDuplicated = ++itBegin; itIndexDuplicated != (*itIndexList).end(); ++itIndexDuplicated)
          _T2F_N3F_V3F_vertexArray[*itIndexDuplicated].normal = N;
      }  
    }
  }
  else
  {
    MFVec3f::const_iterator itNormal = normalFaceArray.begin();
    vector<T2F_N3F_V3F>::iterator itVertex = _T2F_N3F_V3F_vertexArray.begin();
    while (itVertex != _T2F_N3F_V3F_vertexArray.end())
    {
      (*itVertex).normal = *itNormal;
      ++itVertex;
      (*itVertex).normal = *itNormal;
      ++itVertex;
      (*itVertex).normal = *itNormal;
      ++itVertex;
      
      ++itNormal;
    }
  }
}

void IndexedFaceSet::computeColorTexCoordNormals()
{
  MFVec3f normalFaceArray;
  
  if (getNormalPerVertex())
    normalFaceArray.reserve(_indexArray.size()/3);
  
  MFInt32::const_iterator itIndex = _indexArray.begin();
  
  // computing the normals per face.    
  while (itIndex != _indexArray.end())
  {
    SFVec3f A = _T2F_C4F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].vertex; 
    ++itIndex;
    
    SFVec3f B = _T2F_C4F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].vertex;
    ++itIndex;
    
    SFVec3f C = _T2F_C4F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].vertex;
    ++itIndex;
    
    SFVec3f AB = (B - A).normalize();
    SFVec3f AC = (C - A).normalize();
    SFVec3f N;

    if (getCcw())
      N = crossprod(AB, AC);
    else
      N = crossprod(AC, AB);  
    
    normalFaceArray.push_back(N);
  }
  
  MFVec3f::const_iterator itNormal = normalFaceArray.begin();
  if (getNormalPerVertex())
  {
    // adding normals per vertex.  
    itIndex = _indexArray.begin();
    while (itIndex != _indexArray.end())
    {
      _T2F_C4F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].normal += *itNormal;
      ++itIndex;
      _T2F_C4F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].normal += *itNormal;
      ++itIndex;
      _T2F_C4F_N3F_V3F_vertexArray[_X3DtoGLIndex[*itIndex].front()].normal += *itNormal;
      ++itIndex;
      ++itNormal;       
    }
    
    // copying and normalizing to identical values.
    for (vector<MFInt32>::const_iterator itIndexList = _X3DtoGLIndex.begin(); itIndexList != _X3DtoGLIndex.end(); ++itIndexList)
    {
      if (!(*itIndexList).empty())
      {
        _T2F_C4F_N3F_V3F_vertexArray[(*itIndexList).front()].normal.normalize();
        SFVec3f N = _T2F_C4F_N3F_V3F_vertexArray[(*itIndexList).front()].normal;

        MFInt32::const_iterator itBegin = (*itIndexList).begin();
        for (MFInt32::const_iterator itIndexDuplicated = ++itBegin; itIndexDuplicated != (*itIndexList).end(); ++itIndexDuplicated)
          _T2F_C4F_N3F_V3F_vertexArray[*itIndexDuplicated].normal = N;
      }  
    }
  }
  else
  {
    MFVec3f::const_iterator itNormal = normalFaceArray.begin();
    vector<T2F_C4F_N3F_V3F>::iterator itVertex = _T2F_C4F_N3F_V3F_vertexArray.begin();
    while (itVertex != _T2F_C4F_N3F_V3F_vertexArray.end())
    {
      (*itVertex).normal = *itNormal;
      ++itVertex;
      (*itVertex).normal = *itNormal;
      ++itVertex;
      (*itVertex).normal = *itNormal;
      ++itVertex;
      
      ++itNormal;
    }
  }
}

}
}
