/// This file is part of the X3DToolKit library
/// Copyright (C) 2002-2003  Yannick Le Goc (legoc@imag.fr)

/// http://www-imagis.imag.fr/Membres/Yannick.Legoc/X3D/


/// X3DToolKit is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.

/// X3DToolKit is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.

/// You should have received a copy of the GNU General Public License
/// along with X3DToolKit; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
////////////////////////////////////////////////////////////////////////////////
//                            MESH_Mesh.h                                     //
////////////////////////////////////////////////////////////////////////////////

#ifndef MESH_MESH_H
#define MESH_MESH_H

#include "MeshSceneGraphTypes.h"
#include "MESH_TemplateMesh.h"

#include <iostream>
#include <map>
#include <list>

namespace X3DTK {
namespace MESH {

class VertexData;
class EdgeData;
class FaceData;

/// Class providing the content of VertexData for one face.

class VertexData
{
public:
  /// Constructor.
  VertexData();
  /// Destructor.
  ~VertexData();
  
  /// Sets the point.
  void setPoint(const SFPoint3f &point);
  /// Sets the normal of face.
  void setNormalOfFace(const SFVec3f &normal, TemplateFace<FaceData, VertexData, EdgeData> *face);
  /// Sets the normal of face.
  void setColorOfFace(const SFColorRGBA &color, TemplateFace<FaceData, VertexData, EdgeData> *face);
  /// Sets the normal of face.
  void setTexCoordOfFace(const SFVec2f &texCoord, TemplateFace<FaceData, VertexData, EdgeData> *face);
    
  /// Gets the point.
  inline const SFPoint3f &getPoint() const {return _point;};
  /// Gets the normal of face.
  const SFVec3f &getNormalOfFace(TemplateFace<FaceData, VertexData, EdgeData> *face) const;
  /// Sets the normal of face.
  const SFColorRGBA &getColorOfFace(TemplateFace<FaceData, VertexData, EdgeData> *face) const;
  /// Sets the normal of face.
  const SFVec2f &getTexCoordOfFace(TemplateFace<FaceData, VertexData, EdgeData> *face) const;

  /// Gets the list of normals.
  inline const MFVec3f &getNormalArray() const {return _normalArray;};
  /// Gets the list of colors.
  inline const MFColorRGBA &getColorArray() const {return _colorArray;};
  /// Gets the list of texCoords.
  inline const MFVec2f &getTexCoordArray() const {return _texCoordArray;};
  
  /// Gets the map of normal.
  inline std::map<TemplateFace<FaceData, VertexData, EdgeData> *, unsigned short> &getNormalMap() {return _normalMap;};
  /// Gets the map of color.
  inline std::map<TemplateFace<FaceData, VertexData, EdgeData> *, unsigned short> &getColorMap() {return _colorMap;};
  /// Gets the map of texCoord.
  inline std::map<TemplateFace<FaceData, VertexData, EdgeData> *, unsigned short> &getTexCoordMap() {return _texCoordMap;};
    
private:
  SFPoint3f _point;
  MFVec3f _normalArray;
  std::map<TemplateFace<FaceData, VertexData, EdgeData> *, unsigned short> _normalMap;

  MFColorRGBA _colorArray;
  std::map<TemplateFace<FaceData, VertexData, EdgeData> *, unsigned short> _colorMap;

  MFVec2f _texCoordArray;
  std::map<TemplateFace<FaceData, VertexData, EdgeData> *, unsigned short> _texCoordMap;
};

/// Class providing the default content of the template TemplateEdge class.

class EdgeData
{
public:
  /// Constructor.
  EdgeData();
};

/// Class providing the default content of the template TemplateFace class.

class FaceData
{
public:
  /// Constructor.
  FaceData();
  
  /// Sets the normal.
  void setNormal(const SFVec3f &normal);
  /// Sets the color.
  void setColor(const SFColorRGBA &color);
  
  /// Gets the normal.
  inline const SFVec3f &getNormal() const {return _normal;};
  /// Gets the color.
  inline const SFColorRGBA &getColor() const {return _color;};
  
private:
  SFVec3f _normal;
  SFColorRGBA _color;
};

/// Class providing the default content of the TemplateMesh class.

class MeshData
{
public:
  /// Constructor.
  MeshData();
  
  /// Sets the boolean normal, true if vertex has normal.
  void setNormal(bool value);
  /// Sets the boolean color, true if vertex has color.
  void setColor(bool value);
  /// Sets the boolean texCoord, true if vertex has normal.
  void setTexCoord(bool value);
  /// Sets the boolean normalPerVertex.
  void setNormalPerVertex(bool value);
  /// Sets the boolean colorPerVertex.
  void setColorPerVertex(bool value);
  
  /// Returns true if the Mesh has normals.
  inline bool hasNormal() const {return _normal;};
  /// Gets the boolean vertexColor.
  inline bool hasColor() const {return _color;};
  /// Gets the boolean vertexTexCoord.
  inline bool hasTexCoord() const {return _texCoord;};
  /// Gets the boolean faceNormal.
  inline bool getNormalPerVertex() const {return _normalPerVertex;};
  /// Gets the boolean faceColor.
  inline bool getColorPerVertex() const {return _colorPerVertex;};
  
private:
  bool _normal; 
  bool _color; 
  bool _texCoord;
  bool _normalPerVertex;
  bool _colorPerVertex;
};

typedef TemplateMesh<MeshData, VertexData, EdgeData, FaceData> Mesh;

/// Constructor defined to name the node.

template<> 
Mesh::TemplateMesh()
{
  defineTypeName("Mesh");
};

}
}

#endif
