#include "Creator.h"
#include "X3DComponentCreator.h"
#include "Type.h"
#include "Component.h"
#include "CoreCreator.h"
#include "Geometry3DCreator.h"
#include "GroupingCreator.h"
#include "RenderingCreator.h"
#include "ShapeCreator.h"
#include "TexturingCreator.h"
#include "InterpolationCreator.h"
#include "LightingCreator.h"
#include "NURBSCreator.h"

#include <algorithm>
#include <iostream>

using namespace std;

namespace X3DTK {
namespace X3D {

Creator::Creator()
: X3DActor() 
{
  Type::addActor(this);
  
  setComponentCreator(new X3D::CoreCreator());  
  setComponentCreator(new X3D::Geometry3DCreator());  
  setComponentCreator(new X3D::GroupingCreator());  
  setComponentCreator(new X3D::RenderingCreator());  
  setComponentCreator(new X3D::ShapeCreator());  
  setComponentCreator(new X3D::TexturingCreator());  
  setComponentCreator(new X3D::InterpolationCreator());
  setComponentCreator(new X3D::LightingCreator());
  setComponentCreator(new X3D::NURBSCreator());
}

Creator::~Creator()
{
  if (autoDelete)
    for (list<X3DComponentCreator *>::iterator it = componentList_.begin(); it != componentList_.end(); ++it)
    {
      (*it)->removeOneActor();
      if ((*it)->getActorNumber() == 0)
        delete *it;
    }    
  
  Type::removeActor(this);
}

void Creator::setComponentCreator(X3DComponentCreator *component)
{
  //replace the equivalent component.
  X3DComponentCreator *replacedComponent = 0;
  bool replaced = false;
  
  for (list<X3DComponentCreator *>::iterator it = componentList_.begin(); it != componentList_.end(); ++it)
  {
    if ((*it)->getName() == component->getName())
    {
      replacedComponent = *it;
      *it = component; 
      replacedComponent->removeOneActor();
      component->addOneActor();      
      replaced = true;
    }  
  }
  
  if (!replaced)
  {
    componentList_.push_back(component);
    component->addOneActor();
  }
  else
  {
    // removing old functions present in the replacedComponent
    CreationDict componentCreationDict = replacedComponent->getCreationDict();
    for (CreationDict::iterator it = componentCreationDict.begin(); it != componentCreationDict.end(); ++it)
      creationDict_.erase((*it).first); 
  }
  
  //remove replaced component.
  if ((autoDelete) && (replacedComponent != 0) && (replacedComponent->getActorNumber() == 0))
    delete replacedComponent;
    
  //adding new functions
  CreationDict componentCreationDict = component->getCreationDict();
  for (CreationDict::const_iterator it = componentCreationDict.begin(); it != componentCreationDict.end(); ++it)
    creationDict_[(*it).first] = (*it).second;
}

X3D::SFNode Creator::createFromName(const SFString &name) const
{
  CreationDict::const_iterator it = creationDict_.find(name);
  if (it != creationDict_.end())
    return (*it).second->create();
  
  cerr << "warning: " << name << " node not implemented!" << endl;
  return 0;
}

void Creator::reset()
{
  //remove components
  if (autoDelete)
    for (list<X3DComponentCreator *>::iterator it = componentList_.begin(); it != componentList_.end(); ++it)
    {
      (*it)->removeOneActor();
      if ((*it)->getActorNumber() == 0)
        delete *it;
    }    
             
  creationDict_.clear();
  componentList_.clear();
}

SFString Creator::getComponentNameOf(const SFString &name) const
{
  //iterating the list of component.
  for (list<X3DComponentCreator *>::const_iterator it = componentList_.begin(); it != componentList_.end(); ++it)
  {
    if ((*it)->contains(name))
      return (*it)->getName();
  }
  //not found!
  return "";
}

void Creator::addType(const Type *type)
{
  //finding the lowest function in the derivation tree.
  SFString name = type->getName();
  CreationFunction *CF = getCreationFunctionOf(name);
  creationDict_[name] = CF;
}

CreationFunction *Creator::getCreationFunctionOf(const SFString &name) const
{
  CreationFunction *CF = 0;
  
  for (list<X3DComponentCreator *>::const_iterator it = componentList_.begin(); it != componentList_.end(); ++it)
  {
    CF = (*it)->getCreationFunctionOf(name);
    if (CF != 0)
      break;
  }
  
  return CF;
}

Creator *X3DTK::X3D::joinCreator(Creator *N0, Creator *N1)
{
  Creator *N = new Creator();
  
  for (list<X3DComponentCreator *>::const_iterator it = N0->componentList_.begin(); it != N0->componentList_.end(); ++it)
    N->setComponentCreator(*it);
  
  for (list<X3DComponentCreator *>::const_iterator it = N1->componentList_.begin(); it != N1->componentList_.end(); ++it)
    N->setComponentCreator(*it);
  
  return N;  
}

}
}
