#include "IndexedLineSet.h"
#include "X3DColorNode.h"
#include "X3DCoordinateNode.h"

#include <iostream>

using namespace std;

namespace X3DTK {
namespace X3D {

IndexedLineSet::IndexedLineSet()
: X3DGeometryNode(), _color(0), _coord(0), _colorPerVertex(true), _lineWidth(1.0f)
{
  defineTypeName("IndexedLineSet", "Rendering");
}

IndexedLineSet::IndexedLineSet(const X3DColorNode *color, const X3DCoordinateNode *coord, const MFInt32 &colorIndex, SFBool colorPerVertex, const MFInt32 &coordIndex, SFFloat lineWidth)
: X3DGeometryNode(), _color((SFNode)color), _coord((SFNode)coord), _colorIndex(colorIndex), _colorPerVertex(colorPerVertex), _coordIndex(coordIndex), _lineWidth(lineWidth)
{
  defineTypeName("IndexedLineSet", "Rendering");
}

IndexedLineSet::IndexedLineSet(const IndexedLineSet &I)
: X3DGeometryNode(I), _color(I._color), _coord(I._coord), _colorIndex(I._colorIndex), _colorPerVertex(I._colorPerVertex), _coordIndex(I._coordIndex), _lineWidth(I._lineWidth)
{
}  

SFNode IndexedLineSet::clone() const
{
  return new IndexedLineSet(*this);
}

IndexedLineSet::~IndexedLineSet()
{
  removeParentFromChild(this, _color);
  removeParentFromChild(this, _coord);
}  
  
//modifiers
void IndexedLineSet::setColor(const X3DColorNode *color)
{
  removeParentFromChild(this, _color);
  _color = (SFNode)color;
  addParentToChild(this, _color);
}

void IndexedLineSet::setCoord(const X3DCoordinateNode *coord)
{
  removeParentFromChild(this, _coord);
  _coord = (SFNode)coord;
  addParentToChild(this, _coord);
}

void IndexedLineSet::setColorIndex(const MFInt32 &colorIndex)
{
  _colorIndex = colorIndex;
}

void IndexedLineSet::setColorPerVertex(SFBool colorPerVertex)
{
  _colorPerVertex = colorPerVertex;
}

void IndexedLineSet::setCoordIndex(const MFInt32 &coordIndex)
{
  _coordIndex = coordIndex;
  if ((!coordIndex.empty()) && (coordIndex.back() != -1))
    _coordIndex.push_back(-1);
}

void IndexedLineSet::setLineWidth(SFFloat lineWidth)
{
  _lineWidth = lineWidth;
}
  
bool IndexedLineSet::addChild(const SFAbstractNode &N)
{
  if (dynamic_cast<X3DColorNode *>(N) != 0)
  {
    if (_color != 0)
      return false;
      
    setColor(static_cast<X3DColorNode *>(N));
    return true;
  }
  
  if (dynamic_cast<X3DCoordinateNode *>(N) != 0)
  {
    if (_coord != 0)
      return false;
      
    setCoord(static_cast<X3DCoordinateNode *>(N));
    return true;    
  }
  
  cerr << "IndexedLineSet::addChild : a node of type " << N->getTypeName() << " cannot be a child!" << endl;
  return false;
}

bool IndexedLineSet::setChild(const SFAbstractNode &N)
{
  if (dynamic_cast<X3DColorNode *>(N) != 0)
  {
    setColor(static_cast<X3DColorNode *>(N));
    return true;
  }
  
  if (dynamic_cast<X3DCoordinateNode *>(N) != 0)
  {
    setCoord(static_cast<X3DCoordinateNode *>(N));
    return true;    
  }
  
  cerr << "IndexedLineSet::setChild : a node of type " << N->getTypeName() << " cannot be a child!" << endl;
  return false;
}

bool IndexedLineSet::removeChild(const SFAbstractNode &N)
{
  if (_color == N)
  {
    setColor(0);
    return true;
  } 
   
  if (_coord == N)
  {
    setCoord(0);
    return true;
  } 
  
  return false;
}

void IndexedLineSet::load(const X3DFileElement *element)
{ 
  int index;
  index = element->getIndexAttribute("colorIndex");
  if (index != -1)
  {
    _colorIndex = MFInt32(element->getAttribute(index));  
    if ((_colorIndex.empty()) || (_colorIndex.back() != -1))
      _colorIndex.push_back(-1);
  }
  
  index = element->getIndexAttribute("colorPerVertex");
  if (index != -1)
    _colorPerVertex = (element->getAttribute(index).upper() == "TRUE");
  
  index = element->getIndexAttribute("coordIndex");  
  if (index != -1)
  {
    _coordIndex = MFInt32(element->getAttribute(index));
    if ((_coordIndex.empty()) || (_coordIndex.back() != -1))
      _coordIndex.push_back(-1);
  } 
  index = element->getIndexAttribute("lineWidth");    
  if (index != -1)
    _lineWidth = element->getAttribute(index).toFloat();
}

SFString IndexedLineSet::write() const
{
  SFString attr;
  if (!_colorIndex.empty())
    attr += " colorIndex=\"" + toSFString(_colorIndex) + "\"";
  if (!_colorPerVertex)
    attr += " colorPerVertex=\"FALSE\"";
  if (!_coordIndex.empty())  
    attr += " coordIndex=\"" + toSFString(_coordIndex) + "\"";
  if (_lineWidth != 1.0f)
    attr += " lineWidth=\"" + toSFString(_lineWidth) + "\"";
    
  return attr;
}

}
}
